/*****************************************************************************
*   $Id: read.c,v 6.2 1998/07/06 05:17:41 darren Exp $
*
*   Copyright (c) 1996-1998, Darren Hiebert
*
*   This source code is released for free distribution under the terms of the
*   GNU General Public License.
*
*   This module contains low level source and tag file read functions (line
*   splicing and newline conversion for source files are performed at this
*   level).
*****************************************************************************/

/*============================================================================
=   Include files
============================================================================*/
#ifdef HAVE_CONFIG_H
# include <config.h>
#endif
#include <limits.h>		/* to define INT_MAX or MAXINT */

#include "ctags.h"

/*============================================================================
=   Data declarations
============================================================================*/

typedef struct _langTab {
    const char *extension;
    langType language;
} langTab;

/*============================================================================
=   Data definitions
============================================================================*/

sourceFile File = { NULL, NULL, 0, -1, FALSE, 0, FALSE, FALSE };

/*============================================================================
=   Function prototypes
============================================================================*/
static const char *findExtension __ARGS((const char *const fileName));
static boolean isFileHeader __ARGS((const char *const FileName));
static boolean isExtensionInList __ARGS((const char *const extension, const char *const *const list));
static langType getExtensionLanguage __ARGS((const char *const extension));
static langType getFileLanguage __ARGS((const char *const FileName));
static void fileNewline __ARGS((void));
static boolean resizeLineBuffer __ARGS((lineBuf *const pLine));

/*============================================================================
=   Function definitions
============================================================================*/

/*----------------------------------------------------------------------------
 *	Generic line reading with automatic buffer sizing
 *--------------------------------------------------------------------------*/

extern void freeLineBuffer( pLine )
    lineBuf *const pLine;
{
    if (pLine->buffer != NULL)
	free(pLine->buffer);
    pLine->buffer = NULL;
    pLine->size = 0;
}

static boolean resizeLineBuffer( pLine )
    lineBuf *const pLine;
{
    boolean ok = FALSE;

    if (pLine->size <= INT_MAX / 2)
    {
	const int newSize = pLine->size * 2;
	char *const newBuffer = (char *)realloc(pLine->buffer,(size_t)newSize);

	if (newBuffer != NULL)
	{
	    pLine->size = newSize;
	    pLine->buffer = newBuffer;
	    ok = TRUE;
	}
    }
    return ok;
}

/*  Read a newline terminated line from 'fp' and if it overflows the buffer
 *  specified by 'pLine', increase the buffer. The buffer never shrinks. The
 *  size of the buffer will always be the longest line encountered so far.
 */
extern char *readLine( pLine, fp )
    lineBuf *const pLine;
    FILE *const fp;
{
    char *line = NULL;

    if (fp != NULL)		/* to free memory allocated to buffer */
    {
	boolean reReadLine;

	if (pLine->buffer == NULL)	/* if buffer not yet allocated... */
	{
	    pLine->size = 32;
	    pLine->buffer = (char *)malloc((size_t)pLine->size);
	}
	/*  If reading the line places any character other than a null or a
	 *  newline at the last character position in the buffer (one less
	 *  than the buffer size), then we must resize the buffer and
	 *  reattempt to read the line.
	 */
	if (pLine->buffer != NULL) do 
	{
	    const long startOfLine = ftell(fp);
	    char *const pLastChar = pLine->buffer + pLine->size - 2;

	    reReadLine = FALSE;
	    *pLastChar = '\0';
	    line = fgets(pLine->buffer, pLine->size, fp);
	    if (*pLastChar != '\0'  &&  *pLastChar != '\n')	/* overflow */
	    {
		if ((reReadLine = resizeLineBuffer(pLine)))
		    fseek(fp, startOfLine, SEEK_SET);
	    }
	} while (reReadLine);
    }
    return line;
}

/*----------------------------------------------------------------------------
 *	Source file access functions
 *--------------------------------------------------------------------------*/

static const char *findExtension( fileName )
    const char *const fileName;
{
    const char *extension;
    const char *const start = strrchr(fileName, '.');	/* find last '.' */

    if (start == NULL)
	extension = "";
    else
	extension = start + 1;		/* skip to character after '.' */

    return extension;
}

/*  Determines whether the specified file name is considered to be a header
 *  file for the purposes of determining whether enclosed tags are global or
 *  static.
 */
static boolean isFileHeader( fileName )
    const char *const fileName;
{
    const char *const extension = findExtension(fileName);
    boolean header = FALSE;		    /* default unless match found */
    int i;

    for (i = 0 ; Option.headerExt[i] != NULL ; ++i)
    {
	if (strcmp(Option.headerExt[i], extension) == 0)
	{
	    header = TRUE;		    /* found in list */
	    break;
	}
    }
    return header;
}

static boolean isExtensionInList( extension, list )
    const char *const extension;
    const char *const *const list;
{
    boolean isKnown = FALSE;

    if (list != NULL)
    {
	const char *const *pExtension = list;

	while (! isKnown  &&  *pExtension != NULL)
	{
	    if (strcmp(extension, *pExtension) == 0)
		isKnown = TRUE;
	    ++pExtension;
	}
    }
    return isKnown;
}

static langType getExtensionLanguage( extension )
    const char *const extension;
{
    unsigned int i;
    langType language = LANG_IGNORE;

    for (i = 0  ;  i < (int)LANG_COUNT  ;  ++i)
    {
	if (isExtensionInList(extension, Option.langMap[i]))
	{
	    language = (langType)i;
	    break;
	}
    }
    return language;
}

static langType getFileLanguage( fileName )
    const char *const fileName;
{
    const char *const extension = findExtension(fileName);
    langType language;

    if (Option.language != LANG_AUTO)
	language = Option.language;
    else if (isFileHeader(fileName))
	language = LANG_CPP;
    else if (extension[0] == '\0')
	language = LANG_IGNORE;		/* ignore files with no extension */
    else
	language = getExtensionLanguage(extension);

    return language;
}

/*  This function opens a source file, and resets the line counter.  If it
 *  fails, it will display an error message and leave the File.fp set to NULL.
 */
extern boolean fileOpen( fileName )
    const char *const fileName;
{
    boolean opened = FALSE;
    langType language = LANG_IGNORE;

    /*	If another file was already open, then close it.
     */
    if (File.fp != NULL)
    {
	fclose(File.fp);		/* close any open source file */
	File.fp = NULL;
    }

    if (! doesFileExist(fileName))
	error(WARNING | PERROR, "cannot open \"%s\"", fileName);
    else if (! isNormalFile(fileName))
    {
	DebugStatement( if (debug(DEBUG_STATUS))
			  printf("  ignoring %s (special file)\n", fileName); );
    }
    else if ((language = getFileLanguage(fileName)) == LANG_IGNORE)
	DebugStatement( debugOpen(fileName, FALSE, language) );
    else
    {
	File.fp = fopen(fileName, "rb");  /* must be binary mode for fseek() */
	if (File.fp == NULL)
	    error(WARNING | PERROR, "cannot open \"%s\"", fileName);
	else
	{
	    opened = TRUE;

	    File.name	    = fileName;
	    File.lineNumber = 0L;
	    File.seek	    = 0L;
	    File.afterNL    = TRUE;
	    File.warned	    = FALSE;

	    if (strlen(fileName) > TagFile.max.file)
		TagFile.max.file = strlen(fileName);

	    /*	Determine whether this is a header File.
	     */
	    File.isHeader = isFileHeader(fileName);
	    File.language = language;
	    DebugStatement( debugOpen(fileName, File.isHeader, File.language); )
	}
    }
    return opened;
}

extern void fileClose()
{
    if (File.fp != NULL)
    {
	/*  The line count of the file is 1 too big, since it is one-based
	 *  and is incremented upon each newline.
	 */
	if (Option.printTotals)
	    addTotals(0, File.lineNumber - 1L, getFileSize(File.name));

	fclose(File.fp);
	File.fp = NULL;
    }
}

/*  Action to take for each encountered source newline.
 */
static void fileNewline()
{
    File.afterNL = FALSE;
    File.seek	 = ftell(File.fp);
    ++File.lineNumber;
    DebugStatement( if (Option.breakLine == File.lineNumber) lineBreak(); )
}

/*  This function reads a single character from the stream. 
 */
extern int fileGetc()
{
    boolean escaped = FALSE;
    int	c;

    /*	If there is an ungotten character, then return it.  Don't do any
     *	other processing on it, though, because we already did that the
     *	first time it was read.
     */
    if (File.ungetch != '\0')
    {
	c = File.ungetch;
	File.ungetch = '\0';
	return c;	    /* return here to avoid re-calling debugPutc() */
    }

nextChar:	/* not structured, but faster for this critical path */

    /*	If previous character was a newline, then we're starting a line.
     */
    if (File.afterNL)
	fileNewline();

    c = getc(File.fp);
    switch (c)
    {
    default:
	if (escaped)
	{
	    ungetc(c, File.fp);		/* return character after BACKSLASH */
	    c = BACKSLASH;
	}
	break;

    case BACKSLASH:				/* test for line splicing */
	if (escaped)
	    ungetc(c, File.fp);			/* push back one just read */
	else
	{
	    escaped = TRUE;		/* defer test until next character */
	    goto nextChar;
	}
	break;

    /*	The following cases turn line breaks into a canonical form. The three
     *	commonly used forms if line breaks: LF (UNIX), CR (MacIntosh), and
     *	CR-LF (MS-DOS) are converted into a generic newline.
     */
    case CRETURN:
	{
	    const int next = getc(File.fp);	/* is CR followed by LF? */

	    /*	If this is a carriage-return/line-feed pair, treat it as one
	     *	newline, throwing away the line-feed.
	     */
	    if (next != NEWLINE)
		ungetc(next, File.fp);
	}
	c = NEWLINE;				/* convert CR into newline */
    case NEWLINE:
	File.afterNL = TRUE;
	if (escaped)				/* check for line splicing */
	{
	    DebugStatement(
		debugPutc(BACKSLASH, DEBUG_RAW);     /* print the characters */
		debugPutc(c, DEBUG_RAW);	     /*  we're throwing away */
	    )
	    escaped = FALSE;		    /* BACKSLASH now fully processed */
	    goto nextChar;		    /* through away "\NEWLINE" */
	}
	break;
    }

    DebugStatement( debugPutc(c, DEBUG_RAW); )
    return c;
}

extern void fileUngetc( c )
    int c;
{
    File.ungetch = c;
}

/*  Places into the line buffer the contents of the line referenced by
 *  "location".
 */
extern char *getSourceLine( pLine, location )
    lineBuf *const pLine;
    const long location;
{
    const long orignalPosition = ftell(File.fp);
    char *line;

    fseek(File.fp, location, SEEK_SET);
    line = readLine(pLine, File.fp);	
    fseek(File.fp, orignalPosition, SEEK_SET);

    return line;
}

/* vi:set tabstop=8 shiftwidth=4: */
